package ssw.mj.test;

import java.io.StringReader;

import junit.framework.TestCase;

import ssw.mj.ErrorStrings;
import ssw.mj.Scanner;
import ssw.mj.Token;

/**
 * Test cases for the <code>Scanner</code> class.
 * 
 * @author AW
 */
public class ScannerTest extends TestCase {
	private static final String CR = "\r";
	private static final String LF = "\n";
	
	private static final int
		none       = Token.none,
		ident      = Token.ident,
		number     = Token.number,
		charConst  = Token.charConst,
		plus       = Token.plus,
		minus      = Token.minus,
		times      = Token.times,
		slash      = Token.slash,
		rem        = Token.rem,
		eql        = Token.eql,
		neq        = Token.neq,
		gtr        = Token.gtr,
		geq        = Token.geq,
		less       = Token.less,
		leq        = Token.leq,
		and        = Token.and,
		or         = Token.or,
		assign     = Token.assign,
		pplus      = Token.pplus,
		mminus     = Token.mminus,
		semicolon  = Token.semicolon,
		comma      = Token.comma,
		period     = Token.period,
		lpar       = Token.lpar,
		rpar       = Token.rpar,
		lbrack     = Token.lbrack,
		rbrack     = Token.rbrack,
		lbrace     = Token.lbrace,
		rbrace     = Token.rbrace,
		classKW    = Token.classKW,
		continueKW = Token.continueKW,
		doKW       = Token.doKW,
		elseKW     = Token.elseKW,
		finalKW    = Token.finalKW,
		ifKW       = Token.ifKW,
		newKW      = Token.newKW,
		printKW    = Token.printKW,
		readKW     = Token.readKW,
		returnKW   = Token.returnKW,
		voidKW     = Token.voidKW,
		whileKW    = Token.whileKW,
		eof        = Token.eof;

	private static final String
		BIG_NUM           = ErrorStrings.BIG_NUM,
		EMPTY_CHARCONST   = ErrorStrings.EMPTY_CHARCONST,
		INVALID_CHAR      = ErrorStrings.INVALID_CHAR,
		INVALID_CHARCONST = ErrorStrings.INVALID_CHARCONST,
		MISS_SQUOTE       = ErrorStrings.MISS_SQUOTE,
		UNDEF_ESC         = ErrorStrings.UNDEF_ESC;
		
	private static TestPrintWriter output;
	
	// ---------- helpers ----------
	
	private void initScanner (String s) {
		output = new TestPrintWriter();
		Scanner.init(new StringReader(s), output);
	}
	
	private void checkNext (int kind, int line, int col) {
		checkNext(kind, line, col, 0, null);
	}
	
	private void checkNext (int kind, int line, int col, String string) {
		checkNext(kind, line, col, 0, string);
	}
	
	private void checkNext (int kind, int line, int col, int val, String string) {
		Token expected = new Token(kind, line, col, val, string);
		Token actual = Scanner.next();
		assertEquals(expected, actual);
	}

	// ---------------------------------------
	// ---------- test valid inputs ----------
	
	/** Tests all token types. */
	public void testAllTokens () {
		initScanner("anIdentifier 123 'c'" + LF +
		            "+ - * / % == != > >= < <= && || ++ -- ; , . ( ) [ ] { }" + LF +
		            "class continue do else final if new print read return void while" + LF);
		
		checkNext(ident, 1, 1, "anIdentifier");
		checkNext(number, 1, 14, 123, "123");
		checkNext(charConst, 1, 18, 99, null);
		checkNext(plus, 2, 1);
		checkNext(minus, 2, 3);
		checkNext(times, 2, 5);
		checkNext(slash, 2, 7);
		checkNext(rem, 2, 9);
		checkNext(eql, 2, 11);
		checkNext(neq, 2, 14);
		checkNext(gtr, 2, 17);
		checkNext(geq, 2, 19);
		checkNext(less, 2, 22);
		checkNext(leq, 2, 24);
		checkNext(and, 2, 27);
		checkNext(or, 2, 30);
		checkNext(pplus, 2, 33);
		checkNext(mminus, 2, 36);
		checkNext(semicolon, 2, 39);
		checkNext(comma, 2, 41);
		checkNext(period, 2, 43);
		checkNext(lpar, 2, 45);
		checkNext(rpar, 2, 47);
		checkNext(lbrack, 2, 49);
		checkNext(rbrack, 2, 51);
		checkNext(lbrace, 2, 53);
		checkNext(rbrace, 2, 55);
		checkNext(classKW, 3, 1, "class");
		checkNext(continueKW, 3, 7, "continue");
		checkNext(doKW, 3, 16, "do");
		checkNext(elseKW, 3, 19, "else");
		checkNext(finalKW, 3, 24, "final");
		checkNext(ifKW, 3, 30, "if");
		checkNext(newKW, 3, 33, "new");
		checkNext(printKW, 3, 37, "print");
		checkNext(readKW, 3, 43, "read");
		checkNext(returnKW, 3, 48, "return");
		checkNext(voidKW, 3, 55, "void");
		checkNext(whileKW, 3, 60, "while");
		checkNext(eof, 4, 0);

		output.verify();
	}
	
	/** Tests a short program with CR and LF for line separation. */
	public void testCRLFLineSeparators () {
		initScanner("1" + CR + LF +
		            " 2" + CR + LF +
		            "  3 ");
		
		checkNext(number, 1, 1, 1, "1");
		checkNext(number, 2, 2, 2, "2");
		checkNext(number, 3, 3, 3, "3");
		checkNext(eof, 3, 4);
	}

	/** Tests a short program with only LF for line separation. */
	public void testLFLineSeparators() {
		initScanner("1" + LF +
		            " 2" + LF +
		            "  3 ");
		
		checkNext(number, 1, 1, 1, "1");
		checkNext(number, 2, 2, 2, "2");
		checkNext(number, 3, 3, 3, "3");
		checkNext(eof, 3, 4);
	}

	/** Tests a short program with one single line comment. */
	public void testSimpleSingleLineComment () {
		initScanner("class SSLComment { /* Simple single line comment. */ }  ");

		checkNext(classKW, 1, 1, "class");
		checkNext(ident, 1, 7, "SSLComment");
		checkNext(lbrace, 1, 18);
		checkNext(rbrace, 1, 54);
		checkNext(eof, 1, 56);
	}
	
	/** Tests a short program with one multi-line comment. */
	public void testSimpleMultiLineComment () {
		initScanner("class SMLComment {" + LF +
		            "/*\tThis " + LF +
		            "\tis " + LF +
		            "\ta " + LF +
		            "\tsimple " + LF +
		            "\tmulti-line " + LF +
		            "\t comment. */ " + LF +
		            " }");

		checkNext(classKW, 1, 1, "class");
		checkNext(ident, 1, 7, "SMLComment");
		checkNext(lbrace, 1, 18);
		checkNext(rbrace, 8, 2);
		checkNext(eof, 8, 2);
	}

	/** Tests a short program with nested single line comments. */
	public void testNestedSingleLineComment () {
		initScanner("class NSLComment {" + LF +
		            " /* This is a /* nested  /* single-line */ comment. */*/ " + LF +
		            "}");

		checkNext(classKW, 1, 1, "class");
		checkNext(ident, 1, 7, "NSLComment");
		checkNext(lbrace, 1, 18);
		checkNext(rbrace, 3, 1);
		checkNext(eof, 3, 1);		
	}

	/** Tests a short program with nested multi-line comments. */
	public void testNestedMultiLineComment () {
		initScanner("class NMLComment {" + LF +
		            "/*" + LF +
		            "\tThis is a " + LF +
		            "\t/*\tnested  " + LF +
		            "\t\t/* multi-line */" + LF +
		            "\t\tcomment. " + LF +
		            "\t*/" + LF +
		            "*/ " + LF +
		            "}");

		checkNext(classKW, 1, 1, "class");
		checkNext(ident, 1, 7, "NMLComment");
		checkNext(lbrace, 1, 18);
		checkNext(rbrace, 9, 1);
		checkNext(eof, 9, 1);				
	}

  /** Tests some valid character constants. */
	public void testCharConstants () {
		initScanner("space ' ' 'A' 'z' '0' '!' '/' '@' '^' '|' '~'");
		
		checkNext(ident, 1, 1, "space");
		checkNext(charConst, 1, 7, 32, null);	
		checkNext(charConst, 1, 11, 65, null);	
		checkNext(charConst, 1, 15, 122, null);	
		checkNext(charConst, 1, 19, 48, null);	
		checkNext(charConst, 1, 23, 33, null);	
		checkNext(charConst, 1, 27, 47, null);	
		checkNext(charConst, 1, 31, 64, null);	
		checkNext(charConst, 1, 35, 94, null);	
		checkNext(charConst, 1, 39, 124, null);	
		checkNext(charConst, 1, 43, 126, null);	
	}

  /** Tests all valid ESC character sequences. */	
	public void testESCsequences () {
		initScanner("backslash '\\\\'" + LF +
		            "linefeed '\\n'" + LF +
		            "carriage return '\\r'");
		
		checkNext(ident, 1, 1, "backslash");
		checkNext(charConst, 1, 11, 92, null);
		checkNext(ident, 2, 1, "linefeed");
		checkNext(charConst, 2, 10, 10, null);
		checkNext(ident, 3, 1, "carriage");
		checkNext(returnKW, 3, 10, "return");
		checkNext(charConst, 3, 17, 13, null);
		checkNext(eof, 3, 20);
	}
	
	// ----------------------------------------
	// ---------- test invalid input ----------	
	
	/** Tests some invalid characters. */
	public void testInvalidSymbols () {
		initScanner("ch = '';");
		
		checkNext(ident, 1, 1, "ch");
		checkNext(assign, 1, 4);
		checkNext(charConst, 1, 6, 0, null);
		output.addExpectedLine("-- line 1, col 7: " + INVALID_CHARCONST + " ''");
		checkNext(semicolon, 1, 9);
		checkNext(eof, 1, 9);
		
		output.verify();
		
		initScanner("''");
		
		checkNext(charConst, 1, 1, 0, null);
		output.addExpectedLine("-- line 1, col 2: " + INVALID_CHARCONST + " ''");
		checkNext(eof, 1, 3);
		
		output.verify();
		
		initScanner("");
		
		checkNext(none, 1, 1);
		output.addExpectedLine("-- line 1, col 1: " + INVALID_CHAR + " ''");
		checkNext(eof, 1, 1);
		
		output.verify();
	}

	/** Tests an invalid ESC character sequence. */
	public void testInvalidESCsequence () {
		initScanner("unknown '\\x'");
		
		checkNext(ident, 1, 1, "unknown");
		checkNext(charConst, 1, 9, 0, null);
		output.addExpectedLine("-- line 1, col 11: " + UNDEF_ESC + " 'x'");
		checkNext(eof, 1, 12);
		
		output.verify();
	}

	/** Tests character constant without closing single quote. */
	public void testNoClosingQuote () {
		initScanner("'x;");
		
		checkNext(charConst, 1, 1, 'x', null);
		output.addExpectedLine("-- line 1, col 3: " + MISS_SQUOTE);
		checkNext(semicolon, 1, 3);
		checkNext(eof, 1, 3);
		
		output.verify();
	}
	
	/** Tests empty character constant. */
	public void testEmptyCharConst () {
		initScanner("char ch = '';");
		
		checkNext(ident, 1, 1, "char");
		checkNext(ident, 1, 6, "ch");
		checkNext(assign, 1, 9);
		checkNext(charConst, 1, 11, 0, null);
		output.addExpectedLine("-- line 1, col 12: " + EMPTY_CHARCONST);
		checkNext(semicolon, 1, 13);
		checkNext(eof, 1, 13);
		
		output.verify();
	}
	
	/** Tests large and too large integer constants. */
	public void testHugeNumbers () {
		initScanner("int n = 2147483647;" + LF +
		            "int n = 2147483648;" + LF +
		            "int n = 2147483647 + 1;");
		
		checkNext(ident, 1, 1, "int");
		checkNext(ident, 1, 5, "n");
		checkNext(assign, 1, 7);
		checkNext(number, 1, 9, 2147483647, "2147483647");
		checkNext(semicolon, 1, 19);
		checkNext(ident, 2, 1, "int");
		checkNext(ident, 2, 5, "n");
		checkNext(assign, 2, 7);
		checkNext(number, 2, 9, 0, "2147483648");
		output.addExpectedLine("-- line 2, col 19: " + BIG_NUM + ": 2147483648");
		checkNext(semicolon, 2, 19);
		checkNext(ident, 3, 1, "int");
		checkNext(ident, 3, 5, "n");
		checkNext(assign, 3, 7);
		checkNext(number, 3, 9, 2147483647, "2147483647");
		checkNext(plus, 3, 20);
		checkNext(number, 3, 22, 1, "1");
		checkNext(semicolon, 3, 23);
		checkNext(eof, 3, 23);
		
		output.verify();
	}

	// ----------------------------------------------------
	// ---------- test complete correct programs ----------

	/** Tests the shortest valid program: */
	public void testShortestProgram () {
		initScanner("class Shortest { }");

		checkNext(classKW, 1, 1, "class");
		checkNext(ident, 1, 7, "Shortest");
		checkNext(lbrace, 1, 16, null);
		checkNext(rbrace, 1, 18, null);
		checkNext(eof, 1, 18, null);
		
		output.verify();
	}

	/** Tests DoNothing program which already contains some basic constructs.
	 *  This test program uses tabulator characters for indentation.  
	 */
	public void testDoNothingProgram () {
		initScanner("class DoNothing" + LF +
		            "\tint dummy;" + LF +
		            "{" + LF +
		            "\tint doNothing() { dummy=0; return dummy; }" + LF +
		            "" + LF +
		            "\tvoid main() { " + LF +
		            "\t\tdummy= doNothing();" + LF +
		            "\t}" + LF +
		            "}");

		checkNext(classKW, 1, 1, "class");
		checkNext(ident, 1, 7, "DoNothing");
		checkNext(ident, 2, 2, "int");
		checkNext(ident, 2, 6, "dummy");
		checkNext(semicolon, 2, 11);
		checkNext(lbrace, 3, 1);
		checkNext(ident, 4, 2, "int");
		checkNext(ident, 4, 6, "doNothing");
		checkNext(lpar, 4, 15);
		checkNext(rpar, 4, 16);
		checkNext(lbrace, 4, 18);
		checkNext(ident, 4, 20, "dummy");
		checkNext(assign, 4, 25);
		checkNext(number, 4, 26, "0");
		checkNext(semicolon, 4, 27);
		checkNext(returnKW, 4, 29, "return");
		checkNext(ident, 4, 36, "dummy");
		checkNext(semicolon, 4, 41);
		checkNext(rbrace, 4, 43);
		checkNext(voidKW, 6, 2, "void");
		checkNext(ident, 6, 7, "main");
		checkNext(lpar, 6, 11);
		checkNext(rpar, 6, 12);
		checkNext(lbrace, 6, 14);
		checkNext(ident, 7, 3, "dummy");
		checkNext(assign, 7, 8);
		checkNext(ident, 7, 10, "doNothing");
		checkNext(lpar, 7, 19);
		checkNext(rpar, 7, 20);
		checkNext(semicolon, 7, 21);
		checkNext(rbrace, 8, 2);
		checkNext(rbrace, 9, 1);
		checkNext(eof, 9, 1);

		output.verify();
	}

	/** Tests the example program from the lecture script (p. 8, 5.1). 
	 * 	This test program uses space characters for indentation. 
	 */
	public void testScriptExample () {
		initScanner("class P" + LF +
		            "" + LF +
		            "  final int size = 10;" + LF +
		            "" + LF +
		/*line  5*/ "  class Table {" + LF +
		            "    int[] pos;" + LF +
		            "    int[] neg;" + LF +
		            "  }" + LF +
		            "" + LF +
		/*line 10*/ "  Table val;" + LF +
		            "{" + LF +
		            "  void main()" + LF +
		            "    int x, i;" + LF +
		            "  { /*---------- Initialize val ------------*/" + LF +
		/*line 15*/ "    val = new Table;" + LF +
		            "    val.pos = new int[size]; val.neg = new int[size];" + LF +
		            "    i = 0;" + LF +
		            "    while (i < size) {" + LF +
		            "      val.pos[i] = 0; val.neg[i] = 0;" + LF +
		/*line 20*/ "      i++;" + LF +
		            "    }" + LF +
		            "    /*---------- Read values ---------*/" + LF +
		            "    read(x);" + LF +
		            "    while (x > 0) {" + LF +
		/*line 25*/ "      if (0 <= x && x < size) {" + LF +
		            "        val.pos[x]++;" + LF +
		            "      } else if (-size < x && x < 0) {" + LF +
		            "        val.neg[-x]++;" + LF +
		            "      }" + LF +
		/*line 30*/ "      read(x);" + LF +
		            "    }" + LF +
		            "  }" + LF +
		            "}");
		
		checkNext(classKW,1,1,"class"); checkNext(ident,1,7,"P");
		checkNext(finalKW,3,3,"final"); checkNext(ident,3,9,"int"); checkNext(ident,3,13,"size"); checkNext(assign,3,18); checkNext(number,3,20,10,"10"); checkNext(semicolon,3,22);
		checkNext(classKW,5,3,"class"); checkNext(ident,5,9,"Table"); checkNext(lbrace,5,15);
		checkNext(ident,6,5,"int"); checkNext(lbrack,6,8); checkNext(rbrack,6,9); checkNext(ident,6,11,"pos"); checkNext(semicolon,6,14);
		checkNext(ident,7,5,"int"); checkNext(lbrack,7,8); checkNext(rbrack,7,9); checkNext(ident,7,11,"neg"); checkNext(semicolon,7,14);
		checkNext(rbrace,8,3);
		checkNext(ident,10,3,"Table"); checkNext(ident,10,9,"val"); checkNext(semicolon,10,12); 
		checkNext(lbrace,11,1); 
		checkNext(voidKW,12,3,"void"); checkNext(ident,12,8,"main"); checkNext(lpar,12,12); checkNext(rpar,12,13);
		checkNext(ident,13,5,"int"); checkNext(ident,13,9,"x"); checkNext(comma,13,10); checkNext(ident,13,12,"i"); checkNext(semicolon,13,13);
		checkNext(lbrace,14,3);
		checkNext(ident,15,5,"val"); checkNext(assign,15,9); checkNext(newKW,15,11,"new"); checkNext(ident,15,15,"Table"); checkNext(semicolon,15,20);
		checkNext(ident,16,5,"val"); checkNext(period,16,8); checkNext(ident,16,9,"pos"); checkNext(assign,16,13); checkNext(newKW,16,15, "new"); checkNext(ident,16,19,"int"); checkNext(lbrack,16,22); checkNext(ident,16,23,"size"); checkNext(rbrack,16,27); checkNext(semicolon,16,28);
		checkNext(ident,16,30,"val"); checkNext(period,16,33); checkNext(ident,16,34,"neg"); checkNext(assign,16,38); checkNext(newKW,16,40, "new"); checkNext(ident,16,44,"int"); checkNext(lbrack,16,47); checkNext(ident,16,48,"size"); checkNext(rbrack,16,52); checkNext(semicolon,16,53);
		checkNext(ident,17,5,"i"); checkNext(assign,17,7); checkNext(number,17,9,0,"0"); checkNext(semicolon,17,10);
		checkNext(whileKW,18,5,"while"); checkNext(lpar,18,11); checkNext(ident,18,12,"i"); checkNext(less,18,14); checkNext(ident,18,16,"size"); checkNext(rpar,18,20); checkNext(lbrace,18,22);
		checkNext(ident,19,7,"val"); checkNext(period,19,10); checkNext(ident,19,11,"pos"); checkNext(lbrack,19,14); checkNext(ident,19,15,"i"); checkNext(rbrack,19,16); checkNext(assign,19,18); checkNext(number,19,20,0,"0"); checkNext(semicolon,19,21);
		checkNext(ident,19,23,"val"); checkNext(period,19,26); checkNext(ident,19,27,"neg"); checkNext(lbrack,19,30); checkNext(ident,19,31,"i"); checkNext(rbrack,19,32); checkNext(assign,19,34); checkNext(number,19,36,0,"0"); checkNext(semicolon,19,37);
		checkNext(ident,20,7,"i"); checkNext(pplus,20,8); checkNext(semicolon,20,10);
		checkNext(rbrace,21,5);
		checkNext(readKW,23,5,"read"); checkNext(lpar,23,9); checkNext(ident,23,10,"x"); checkNext(rpar,23,11); checkNext(semicolon,23,12);
		checkNext(whileKW,24,5,"while"); checkNext(lpar,24,11); checkNext(ident,24,12,"x"); checkNext(gtr,24,14); checkNext(number,24,16,0,"0"); checkNext(rpar,24,17); checkNext(lbrace,24,19);
		checkNext(ifKW,25,7,"if"); checkNext(lpar,25,10); checkNext(number,25,11,0,"0"); checkNext(leq,25,13); checkNext(ident,25,16,"x"); checkNext(and,25,18); checkNext(ident,25,21,"x"); checkNext(less,25,23); checkNext(ident,25,25,"size"); checkNext(rpar,25,29); checkNext(lbrace,25,31);
		checkNext(ident,26,9,"val"); checkNext(period,26,12); checkNext(ident,26,13,"pos"); checkNext(lbrack,26,16); checkNext(ident,26,17,"x"); checkNext(rbrack,26,18); checkNext(pplus,26,19); checkNext(semicolon,26,21);
		checkNext(rbrace,27,7); checkNext(elseKW,27,9,"else"); checkNext(ifKW,27,14,"if"); checkNext(lpar,27,17); checkNext(minus,27,18); checkNext(ident,27,19,"size"); checkNext(less,27,24); checkNext(ident,27,26,"x"); checkNext(and,27,28); checkNext(ident,27,31,"x"); checkNext(less,27,33); checkNext(number,27,35,0,"0"); checkNext(rpar,27,36); checkNext(lbrace,27,38);
		checkNext(ident,28,9,"val"); checkNext(period,28,12); checkNext(ident,28,13,"neg"); checkNext(lbrack,28,16); checkNext(minus,28,17); checkNext(ident,28,18,"x"); checkNext(rbrack,28,19); checkNext(pplus,28,20); checkNext(semicolon,28,22);
		checkNext(rbrace,29,7);
		checkNext(readKW,30,7,"read"); checkNext(lpar,30,11); checkNext(ident,30,12,"x"); checkNext(rpar,30,13); checkNext(semicolon,30,14);
		checkNext(rbrace,31,5);
		checkNext(rbrace,32,3);
		checkNext(rbrace,33,1);
		checkNext(eof,33,1);

		output.verify();
	}	

	/** Starts a text-mode test runner that runs all scanner tests. */	
	public static void main (String[] args) {
		junit.textui.TestRunner.run(ScannerTest.class);	
	}
}
